/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

This software is based in part on the work of the Independent JPEG Group.
Source code has been derived from the libjpeg archive, a separate package
copyright to Thomas G. Lane.  Libjpeg is publicly available on terms that
are not related to this Package.  The original libjpeg source code can be
obtained from http://www.ijg.org.

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/assign.h>
#include <pandora/modules/filesystem.h>

#include "lib/jpeglib.h"
#include "lib/jerror.h"
#include "lib/jpegint.h"

#define VER_JPEGMod 1.0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_JPEGMod, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2001.  All rights reserved.",
   "December 2001",
   "JPEG"
};

struct KernelBase *KernelBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR JPEGClass = 0;
static OBJECTPTR modFileSystem = 0;

static ERROR JPEG_Activate(struct Picture *, APTR);
static ERROR JPEG_Init(struct Picture *, APTR);
static ERROR JPEG_Query(struct Picture *, APTR);
static ERROR JPEG_SaveImage(struct Picture *, struct acSaveImage *);
static ERROR JPEG_SaveToObject(struct Picture *, struct acSaveToObject *);

static struct ActionArray JPEGActions[] = {
   { AC_Activate,     JPEG_Activate },
   { AC_Init,         JPEG_Init },
   { AC_Query,        JPEG_Query },
   { AC_SaveImage,    JPEG_SaveImage },
   { AC_SaveToObject, JPEG_SaveToObject },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** File System module ***/

   if (NewPrivateObject(ID_MODULE, NULL, &modFileSystem) IS ERR_Okay) {
      SetField(modFileSystem, FID_Name, FT_STRING, "filesystem");
      if (Action(AC_Init, modFileSystem, NULL) IS ERR_Okay) {
         GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase);
      }
      else return(ObjectError(argModule, ERH_InitModule, ERR_Init));
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_NewObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&JPEGClass, NULL,
      FID_BaseClassID|TLONG,       ID_PICTURE,
      FID_SubClassID|TLONG,        ID_JPEG,
      FID_Name|TSTRING,            "JPEG",
      FID_Category|TLONG,          CCF_GRAPHICS,
      FID_FileExtension|TSTRING,   "jpeg;jpeg;jfif",
      FID_FileDescription|TSTRING, "JPEG Picture",
      FID_Actions|TPTR,            JPEGActions,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = 0; }
   if (JPEGClass)     { acFree(JPEGClass); JPEGClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Picture: Activate
*/

static ERROR JPEG_Activate(struct Picture *Self, APTR Void)
{
   struct Bitmap *Bitmap;
   struct mtDrawRGBPixel RGBArgs;
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   struct RGB rgb;
   JSAMPARRAY buffer;
   JSAMPROW row;
   JSAMPARRAY colormap = cinfo.colormap;
   LONG   row_stride;
   STRING location;
   WORD   x, y, i;

   /*** Return if the picture object has already been activated ***/

   if (Self->Bitmap->Head.ObjectFlags & NF_INITIALISED) return(ERR_Okay);

   if (!Self->prvFile) {
      if (GetField(Self, FID_Location, FT_STRING, &location) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_GetField));

      if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
         FID_Location|TSTRING, location,
         FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
         TAGEND) != ERR_Okay) {
         DPrintF("!Init:","[JPEG:%d] Failed to open file \"%s\".", Self->Head.UniqueID, location);
         return(ERR_File);
      }
   }

   /*** Read the JPEG file ***/

   ActionTags(AC_Seek, Self->prvFile, 0.0, SEEK_START);

   Bitmap = Self->Bitmap;
   cinfo.err = jpeg_std_error((struct jpeg_error_mgr *)&jerr);
   jpeg_create_decompress(&cinfo);
   jpeg_stdio_src(&cinfo, Self->prvFile);
   jpeg_read_header(&cinfo, TRUE);

   if (!Bitmap->Width)        Bitmap->Width       = cinfo.image_width;
   if (!Bitmap->Height)       Bitmap->Height      = cinfo.image_height;
   if (!Self->DisplayWidth)   Self->DisplayWidth  = Bitmap->Width;
   if (!Self->DisplayHeight)  Self->DisplayHeight = Bitmap->Height;
   if (!Bitmap->Type)         Bitmap->Type        = BMP_CHUNKY;
   if (!Bitmap->BitsPerPixel) Bitmap->BitsPerPixel = 24;

   if (Action(AC_Query, Self->Bitmap, NULL) IS ERR_Okay) {
      if (Action(AC_Init, Self->Bitmap, NULL) != ERR_Okay) {
         jpeg_destroy_decompress(&cinfo);
         return(ERR_Init);
      }
   }
   else {
      jpeg_destroy_decompress(&cinfo);
      return(ERR_Query);
   }

   if (Self->Flags & IMG_RESIZEX) cinfo.output_width = Bitmap->Width;
   if (Self->Flags & IMG_RESIZEY) cinfo.output_height = Bitmap->Height;

   DPrintF("3Init:","[JPEG] Starting JPEG decompression, buffer size %dx%d.", Bitmap->Width, Bitmap->Height);

   jpeg_start_decompress(&cinfo);

   /*** Decompression routine ***/

   DPrintF("3Init:","[JPEG] Unpacking data to a %dbpp Bitmap...", Bitmap->BitsPerPixel);

   row_stride = cinfo.output_width * cinfo.output_components;
   buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr)&cinfo, JPOOL_IMAGE, row_stride, 1);
   RGBArgs.YCoord = 0;
   for (y=0; cinfo.output_scanline < cinfo.output_height; y++) {
      jpeg_read_scanlines(&cinfo, buffer, 1);
      row = buffer[0];

      if (cinfo.output_components IS 3) {
         for (x=0; x < cinfo.output_width; x++) {
            rgb.Red   = *row++;
            rgb.Green = *row++;
            rgb.Blue  = *row++;
            rgb.Alpha = 0;
            Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);
         }
      }
      else if (cinfo.out_color_space IS JCS_RGB) {
         for (x=0; x < cinfo.output_width; x++) {
            i = *row++;
            rgb.Red   = GETJSAMPLE(colormap[0][i]);
            rgb.Green = GETJSAMPLE(colormap[1][i]);
            rgb.Blue  = GETJSAMPLE(colormap[2][i]);
            rgb.Alpha = 0;
            Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);
         }
      }
      else {
         /*** Greyscale ***/
         for (x=0; x < cinfo.output_width; x++) {
            rgb.Red = rgb.Green = rgb.Blue = *row++;
            rgb.Alpha = 0;
            Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);
         }
      }

      RGBArgs.YCoord++;
   }

   DPrintF("4Init:","[JPEG] Decompression complete.");
   jpeg_finish_decompress(&cinfo);
   jpeg_destroy_decompress(&cinfo);

   acFree(Self->prvFile);
   Self->prvFile = NULL;

   if (Self->Flags & PCF_CACHE) Action(MT_CacheBitmap, Self, NULL);

   return(ERR_Okay);
}

/*****************************************************************************
** Picture: Init
*/

static ERROR JPEG_Init(struct Picture *Self, APTR Void)
{
   UBYTE *buffer;
   STRING location;

   GetField(Self, FID_Location, FT_STRING, &location);

   if ((!location) OR (Self->Flags & IMG_NEW)) {
      /* If no location has been specified, assume that the picture is being
      ** created from scratch (e.g. to save an image to disk).  The
      ** programmer is required to specify the dimensions and colours of the
      ** Bitmap so that we can initialise it.
      */

      if (!Self->Bitmap->Width) Self->Bitmap->Width = Self->DisplayWidth;
      if (!Self->Bitmap->Height) Self->Bitmap->Height = Self->DisplayHeight;

      if ((Self->Bitmap->Width) AND (Self->Bitmap->Height)) {
         if (Action(AC_Init, Self->Bitmap, NULL) IS ERR_Okay) {
            return(ERR_Okay);
         }
         else return(ObjectError(Self, ERH_Init, ERR_Init));
      }
      else return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));
   }
   else if (GetField(Self, FID_Header, FT_POINTER, &buffer) IS ERR_Okay) {
      if ((buffer[0] IS 0xff) AND (buffer[1] IS 0xd8) AND (buffer[2] IS 0xff) AND
          ((buffer[3] IS 0xe0) OR (buffer[3] IS 0xe1) OR (buffer[3] IS 0xfe))) {
         DPrintF("JPEG:","The file is a JPEG picture.");
         if (!(Self->Flags & IMG_QUERY)) acActivate(Self);
         return(ERR_Okay);
      }
      else DPrintF("JPEG:","The file is not a JPEG picture.");
   }

   return(ERR_NoSupport);
}

/*****************************************************************************
** Picture: Query
*/

static ERROR JPEG_Query(struct Picture *Self, APTR Void)
{
   struct jpeg_decompress_struct *cinfo;
   struct jpeg_error_mgr jerr;
   struct Bitmap *Bitmap;
   STRING location;

   DPrintF("Query()","[JPEG:%d]", Self->Head.UniqueID);

   if (!Self->prvFile) {
      if (GetFields(Self, FID_Location|TSTRING, &location, TAGEND) != ERR_Okay) return(ObjectError(Self, ERH_Query, ERR_GetField));

      if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
            FID_Location|TPTR, location,
            FID_Flags|TLONG,   FL_READ|FL_APPROXIMATE,
            TAGEND) != ERR_Okay) return(ObjectError(Self, ERH_Query, ERR_CreateObject));
   }

   ActionTags(AC_Seek, Self->prvFile, 0.0, SEEK_START);
   if (AllocMemory(sizeof(struct jpeg_decompress_struct), MEM_DATA, (APTR *)&cinfo, NULL) IS ERR_Okay) {
      Bitmap = Self->Bitmap;
      cinfo->err = jpeg_std_error((struct jpeg_error_mgr *)&jerr);
      jpeg_create_decompress(cinfo);
      jpeg_stdio_src(cinfo, Self->prvFile);
      jpeg_read_header(cinfo, FALSE);

      if (!Bitmap->Width)        Bitmap->Width        = cinfo->image_width;
      if (!Bitmap->Height)       Bitmap->Height       = cinfo->image_height;
      if (!Self->DisplayWidth)   Self->DisplayWidth   = Bitmap->Width;
      if (!Self->DisplayHeight)  Self->DisplayHeight  = Bitmap->Height;
      if (!Bitmap->Type)         Bitmap->Type         = BMP_CHUNKY;
      if (!Bitmap->BitsPerPixel) {
         Bitmap->BitsPerPixel = 24;
         Bitmap->BytesPerPixel = 3;
      }

      jpeg_destroy_decompress(cinfo);
      FreeMemory(cinfo);

      return(Action(AC_Query, Bitmap, NULL));
   }
   else return(ObjectError(Self, ERH_Query, ERR_Memory));
}

/*****************************************************************************
** Picture: SaveImage
*/

static ERROR JPEG_SaveImage(struct Picture *Self, struct acSaveImage *Args)
{
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   struct Bitmap *bitmap;
   struct RGB rgb;
   OBJECTPTR file;
   ERROR error;
   STRING location;
   LONG y, x;
   WORD index;
   JSAMPROW row_pointer[1];

   DPrintF("~SaveImage()","[JPEG:%d]", Self->Head.UniqueID);

   /*** Initialise variables ***/

   bitmap    = Self->Bitmap;
   file      = NULL;
   error     = ERR_Failed;

   /*** Open the data file ***/

   if ((Args) AND (Args->DestinationID)) {
      if (AccessObject(Args->DestinationID, 3000, &file) != ERR_Okay) {
         DPrintF("@SaveImage:","Failed to access destination object #%d.", Args->DestinationID);
         return(ERR_AccessObject);
      }
   }
   else {
      if (GetFields(Self, FID_Location|TSTRING, &location, TAGEND) != ERR_Okay) {
         ObjectError(Self, ERH_SaveImage, ERR_MissingLocation);
         StepBack();
         return(ERR_GetField);
      }

      if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Location|TSTRING, location,
            FID_Flags|TLONG,      FL_NEW|FL_WRITE,
            TAGEND) != ERR_Okay) {
         ObjectError(Self, ERH_SaveImage, ERR_CreateObject);
         StepBack();
         return(ERR_CreateObject);
      }
   }

   /*** Allocate jpeg structures ***/

   cinfo.err = jpeg_std_error((struct jpeg_error_mgr *)&jerr);
   jpeg_create_compress(&cinfo);
   jpeg_stdio_dest(&cinfo, file);

   cinfo.image_width      = Self->Bitmap->Width; 	/* image width and height, in pixels */
   cinfo.image_height     = Self->Bitmap->Height;
   cinfo.input_components = 3;	/* # of color components per pixel */
   cinfo.in_color_space   = JCS_RGB; /* colorspace of input image */
   jpeg_set_defaults(&cinfo);

   jpeg_set_quality(&cinfo, (Self->Quality * 255) / 100, TRUE); /* Quality is between 1 and 255 (lower is better) */

   jpeg_start_compress(&cinfo, TRUE);

   {
      UBYTE buffer[3 * Self->Bitmap->Width];

      for (y=0; y < Self->Bitmap->Height; y++) {
         row_pointer[0] = buffer;
         index = 0;
         for (x=0; x < Self->Bitmap->Width; x++) {
            Self->Bitmap->ReadUCRPixel(Self->Bitmap, x, y, &rgb);
            buffer[index++] = rgb.Red;
            buffer[index++] = rgb.Green;
            buffer[index++] = rgb.Blue;
         }
         jpeg_write_scanlines(&cinfo, row_pointer, 1);
      }
   }

   jpeg_finish_compress(&cinfo);

   jpeg_destroy_compress(&cinfo);

   if (file) {
      if ((Args) AND (Args->DestinationID)) ReleaseObject(file);
      else Action(AC_Free, file, NULL);
   }

   StepBack();
   return(ERR_Okay);
}


/*****************************************************************************
** Picture: SaveToObject
*/

static ERROR JPEG_SaveToObject(struct Picture *Self, struct acSaveToObject *Args)
{
   /*** Pass the action call directly on to SaveImage ***/
   return(ActionTags(AC_SaveImage, Self, Args->DestinationID, NULL));
}
